#!bash Coloured prompts.
# profile-required: TERM.bashrc
#
# The prompt comprises multiple parts, some of which may be hidden by unsetting 
# shell variables or using the ``prompt'' function.
#
# The first part of the prompt is user@host where host is highlighted in 
# green if the last command exited 0 or in red otherwise.
# This part will be shown only if __ps1_user is 1.  By default it is 1.
# The success and failure colours can be changed by modifying 
# $PROMPT_OK_COLOUR and $PROMPT_FAILED_COLOUR respectively.
#
# The second part of the prompt is taken from git-completion.bashrc.
# It is shown only if __ps1_git is 1.  By default it is 0.
#
# The third part of the prompt is taken from p4-completion.bashrc.
# It is shown only if __ps1_p4 is 1.  By default it is 0.
#
# The fourth part of the prompt is taken from svn-completion.bashrc.
# It is shown only if __ps1_svn is 1.  By default it is 0.
#
# The fifth part of the prompt is the exit status of the last command.
# This part will be shown only if __ps1_user is set and the exit status is 
# non-zero.
#
# The sixth part of the prompt is the number of background jobs managed
# by the shell, in square brackets.  If all background jobs are running,
# their number will be shown as [n].  If some are stopped, the number of
# running (r) and total (t) jobs will be shown as [r/t].
# This part will be shown only if __ps1_bg is 1.  By default it is 0.

# The final part of the prompt is the (full) working directory and $ string.
# If the shell is running as root the # string's colour can be changed by
# modifying $ROOT_OK_COLOUR and $ROOT_FAILED_COLOUR.
#
# Colouring is performed by the __ps1_col() and __ps1_ret() functions.
# We redirect stderr to /dev/null when calling these functions to prevent 
# bash complaining about not knowing them when you su to another user, 
# retaining PS1 but not the function definitions.
#
# Note that $? is passed as an argument to - and is returned from - all 
# functions.  As $? is set following any shell activity it is only guaranteed 
# to represent the return code of the last command at the beginning of __ps1().
# By passing between subsequent functions we ensure that it is available for 
# __ps1_ret().
#

# Pick a colour based on the terminal capabilities.
# OK: dark green.
# Failed: dark red.
# Git: royal blue.
# P4: yellow.
# SVN: magenta.
case $(tput colors) in
  256)
    PROMPT_BACKGROUND_COLOUR="0;48;5;26"
    PROMPT_OK_COLOUR="1;38;5;34"
    PROMPT_FAILED_COLOUR="1;38;5;160"
    ROOT_OK_COLOUR="0"
    ROOT_FAILED_COLOUR="0"
    GIT_COLOUR="0;38;5;33"
    SVN_COLOUR="0;38;5;127"
    P4_COLOUR="0;38;5;142"
  ;;

  88)
    PROMPT_BACKGROUND_COLOUR="0;48;5;18"
    PROMPT_OK_COLOUR="1;38;5;24"
    PROMPT_FAILED_COLOUR="1;38;5;48"
    ROOT_OK_COLOUR="0"
    ROOT_FAILED_COLOUR="0"
    GIT_COLOUR="0;38;5;23"
    SVN_COLOUR="0;38;5;49"
    P4_COLOUR="0;38;5;56"
  ;;

  *)
    PROMPT_BACKGROUND_COLOUR="0;44"
    PROMPT_OK_COLOUR="1;32"
    PROMPT_FAILED_COLOUR="1;31"
    ROOT_OK_COLOUR="0"
    ROOT_FAILED_COLOUR="0"
    GIT_COLOUR="0;36"
    SVN_COLOUR="0;35"
    P4_COLOUR="0;33"
  ;;
esac

function __ps1() {
  # Default __ps1_user to 1.
  [ -z "$__ps1_user" ] && __ps1_user=1

  PS1='\[\033[$(__ps1_background $?)m\]$(__ps1_user $? \u@)\[\033[$(__ps1_col $? 2>/dev/null)m\]$(__ps1_user $? \h)\[\033[$(__ps1_colour_escape $? $GIT_COLOUR $PROMPT_BACKGROUND_COLOUR)m\]$(__ps1_git $? 2>/dev/null)\[\033[0m\]\[\033[$(__ps1_colour_escape $? $P4_COLOUR $PROMPT_BACKGROUND_COLOUR)m\]$(__ps1_p4 $? 2>/dev/null)\[\033[0m\]\[\033[$(__ps1_colour_escape $? $SVN_COLOUR $PROMPT_BACKGROUND_COLOUR)m\]$(__ps1_svn $? 2>/dev/null)\[\033[$(__ps1_background $?)m\]$(__ps1_ret $? 2>/dev/null)$(__ps1_bg $?)$(__ps1_colon $?)$(__ps1_short $?)\[\033[$(__ps1_root $? 2>/dev/null)m\]\$\[\033[0m\] '
  return 0
}

# iTerm doesn't like it if you set bold and colour at the same time.
function __ps1_colour_escape() {
  local ret=$1; shift
  local bold="${1%%;*}"
  local colour="${1#*;}"
  local bgcolour="${2#*;}"

  echo -en "${bold}m\033[$colour"
  [ "$__ps1_background" = 1 ] && echo -en "m\033[$bgcolour"
  return $ret
}

function __ps1_background() {
  local ret=$1; shift
  echo -n "0"
  [ "$__ps1_background" = 1 ] && echo -en "m\033[${PROMPT_BACKGROUND_COLOUR}"
  return $ret
}

function __ps1_col() {
  local ret=$1; shift
  if [ $ret -gt 0 ]; then
    __ps1_colour_escape $ret "$PROMPT_FAILED_COLOUR" "$PROMPT_BACKGROUND_COLOUR"
  else
    __ps1_colour_escape $ret "$PROMPT_OK_COLOUR" "$PROMPT_BACKGROUND_COLOUR"
  fi
  return $ret
}

function __ps1_root() {
  local ret=$1; shift
  if [ $ret -gt 0 ]; then
    __ps1_colour_escape $ret "$ROOT_FAILED_COLOUR" "$PROMPT_BACKGROUND_COLOUR"
  else
    __ps1_colour_escape $ret "$ROOT_OK_COLOUR" "$PROMPT_BACKGROUND_COLOUR"
  fi
  return $ret
}

function __ps1_ret() {
  [ "$__ps1_user" = "1" ] || return $1
  [ $1 -gt 0 ] && echo -n " ($1)"
  return $1
}

function __ps1_user() {
  local ret=$1; shift
  [ "$__ps1_user" = "1" ] || return $ret
  echo -n ${1+"$@"}
  return $ret
}

function __ps1_git() {
  [ "$__ps1_git" = "1" ] || return $1
  if [ "$__ps1_user" = "1" ]; then
    __git_ps1 ' %s'
  else
    __git_ps1 '%s'
  fi
  return $1
}

function __ps1_p4() {
  [ "$__ps1_p4" = "1" ] || return $1
  if [ "$__ps1_user" = "1" -o "$__ps1_git" = "1" ]; then
    __p4_ps1 ' %s'
  else
    __p4_ps1 '%s'
  fi
  return $1
}

function __ps1_svn() {
  [ "$__ps1_svn" = "1" ] || return $1
  if [ "$__ps1_user" = "1" -o "$__ps1_git" = "1" -o "$__ps1_p4" = "1" ]; then
    __svn_ps1 ' %s'
  else
    __svn_ps1 '%s'
  fi
  return $1
}

function __ps1_bg() {
  [ "$__ps1_bg" = "1" ] || return $1
  local job
  local running=0; for job in $(builtin jobs -pr); do running=$((running+1)); done
  local total=0; for job in $(builtin jobs -p); do total=$((total+1)); done
  [ $total = 0 ] && return $1
  if [ -z "$2" ]; then
    [ "$__ps1_user" = "1" ] && echo -n " "
    echo -n "["
    [ $running = $total ] || echo -n "$running/"
    echo -n "$total]"
  else
    echo $2
  fi
  return $1
}

function __ps1_colon() {
  local all="$__ps1_user$__ps1_git$__ps1_p4$__ps1_svn$(__ps1_bg $1 1)"
  [ "${all/1/}" = "$all" ] || echo -n ":"
  return $1
}

function __ps1_short() {
  local home=${HOME%%/}
  local pwd=${PWD/#$home/\~}
  local dirtrim=${PROMPT_DIRTRIM//[^0-9]/}

  if [ "${dirtrim:0:1}" = "0" ]; then
    echo "$pwd"
    return $1
  fi

  dirtrim=${dirtrim##0}
  if [ -z "$dirtrim" ]; then
    local prompt="$USER$HOSTNAME$pwd"
    local width=$(((COLUMNS*2)/3))
    if [ ${#prompt} -le ${width:-53} ]; then
      echo "$pwd"
      return $1
    else
      dirtrim=1
    fi
  fi

  local dirname=${pwd##*/}
  local basename=${pwd%/$dirname}
  local reversed=
  local component
  for component in ${basename//\// }; do
    reversed="$component $reversed"
  done
  local n=1
  local short=
  for component in $reversed; do
    [ $n = 1 -a "$PWD" = "$pwd" ] || short="/$short"
    if [ $n -ge $dirtrim ]; then
      short="${component:0:1}$short"
    else
      short="$component$short"
    fi
    n=$((n+1))
  done

  echo "$short$dirname"
  return $1
}

function prompt() {
  local blurb="Usage: prompt hide|show <what>"

  if [ $# -lt 2 ]; then
    echo >&2 "$blurb"
    return 1
  fi

  action="$1"
  if [ ! "$action" = "hide" -a ! "$action" = "show" ]; then
    echo >&2 "$blurb"
    return 1
  fi
  if [ "$action" = "hide" ]; then
    action=0
  else
    action=1
  fi

  what="$(echo $2 | env LANG= LC_ALL= LC_CTYPE= tr '[:upper:]' '[:lower:]')"
  eval __ps1_$what=$action
}

__ps1
